# encoding: utf-8

require 'yaml'

require File.join(__dir__, 'configurates', 'provider')
require File.join(__dir__, 'configurates', 'provider', 'yaml')
require File.join(__dir__, 'configurates', 'provider', 'environment')
require File.join(__dir__, 'configurates', 'storage')

#
# Settings::Config is a module that provides agile way to work with
# configuration, which can be useful for object initialization with default
# settings. See details below:
#
# class MyClass
#
#   def initialize
#     config = Settings::Config.create do
#       merge_storage( Settings::Yaml.load('path/to/my/yaml/file'),
#                      root: 'myclass/default' )
#     end
#
#     if config.download?
#       @power = config.power
#       @value = config.actual.value
#     else
#       print config['WARNING_MESSAGE']
#     end
#   end
# end
#
# This code demonstrates next concepts:
#
# First. You can choose what config part should be loaded. It can be useful for
# small projects where you do not want to create more than one config file. It
# means even if whole config placed in one place it is possible to load small
# data/file part.
#
# Second. You can select an object from where configs can be loaded. By default
# it can be Settings::Yaml or Settings::ENV provider
#
# Third. Chains are allowed.
#
#
module Settings
  #
  #
  #
  class Config
    attr_reader :root

    include Provider

    #
    # Currently provider(s) can be registered only when we creating new instance
    # based on Settings:Config class. It means self.create should be a higher
    # function.
    #
    def self.create(&callback)
      msg = 'Provider not being chosen'
      raise NoBlockGiven, msg unless block_given?
      new(&callback)
    end

    private

    #
    # Run &callback code in object instance.
    #
    def initialize(&callback)
      instance_eval(&callback)
    end

    #
    # Provide message redirect in the @root (storage) if it's possible.
    #
    def method_missing(msg, *args, &blk)
      message_for_root?(msg) ? @root.send(msg, *args, &blk) : super
    end

    #
    # And extend respond_to with storate's responses
    #
    def respond_to?(msg, include_private = false)
      super || message_for_root?(msg) ? true : false
    end

    #
    #
    #
    def message_for_root?(msg)
      @root.send(:respond_to?, msg) ? true : false
    end

    #
    # This is supposed to be raised when we trying to create instance for
    # Settings::Config class without &codeblock
    #
    class NoBlockGiven < StandardError; end
  end # class Settings::Config
end # module Settings
