# encoding: utf-8

module Settings
  #
  # Module Provider is an inclusion for Settings::Config class.
  #
  # This module used for provider validations and data normalization.
  #
  # Details:
  #
  # Any provider (f.e. Settings::Yaml should respond to :data and :load
  # methods, otherwise an exception will be raised.
  #
  # Any provider after provider.data method call should return non-empty hash,
  # otherwise an exception will be raised.
  #
  # Provider should present non-empty sub-hash if optional parameter "#{root}"
  # was used. For example for root: 'data/day1', original hash needs to include
  # subhash like:
  #
  # hash = {…, …, data: {…, day1: { …, … }, … }, … }
  #
  module Provider
    private

    #
    # Can be called only when Settings::Config class is initializing.
    #
    def merge_storage(provider, root: nil)
      msg = "Provider #{provider} does not respond on :data or :load messages"
      raise ProviderNotRespond, msg unless provider.class.respond_to?(:load) &&
                                           provider.respond_to?(:data)

      hsh = validate_and_extract_hash_from(provider)
      extract_sub_hash_from(hsh, root)
      defined?(@root) ? @root.merge(hsh) : @root = Storage.create(hsh)
    end

    #
    # mutate hash unless root parameter is nil. Raise an exception if
    # mutation result is not a valid hash.
    #
    def extract_sub_hash_from(hsh, root)
      return hsh if root.nil?
      tmp = hsh
      root.split('/').each do |path|
        tmp = tmp[path.to_sym]
        msg = "No sub hash with symbol name '#{path}' for #{hsh} was found!"
        raise RootNotExist, msg unless hash_valid?(tmp)
      end
      hsh.replace tmp
    end

    #
    # encapsulate & extract & validate hash from provider
    #
    def validate_and_extract_hash_from(provider)
      hsh = provider.data.dup
      msg = "#{hsh} is not valid or empty Hash!"
      raise HashNotValid, msg unless hash_valid?(hsh)
      hsh
    end

    #
    # valid hash is a non-empty hash object
    #
    def hash_valid?(hsh)
      hsh.is_a?(Hash) && !hsh.empty? ? true : false
    end

    #
    # This is supposed to be raised when provider not respond to :data
    # or :load messages
    #
    class ProviderNotRespond < StandardError; end

    #
    # This is supposed to be raised when hash is not really a Hash or just
    # when it is empty.
    #
    class HashNotValid < StandardError; end

    #
    # This is supposed to be raised when we cant extract a valid sub-hash
    # from hash. In other words Root for hash not exist.
    #
    class RootNotExist < StandardError; end
  end # module Settings::Provider
end # module Settings
